@@ -29,8 +29,10 @@ import fcntl
 import fnmatch
 import gettext
 import grp
+import io
 import locale
 import logging
+import logging.handlers
 import re
 import os
 import select
@@ -40,8 +42,9 @@ import subprocess
 import sys
 
 
+from datetime import date
 from email.message import Message
-from gettext import lgettext as _
+from gettext import gettext as _
 from io import StringIO
 from optparse import (
     OptionParser,
@@ -440,7 +443,7 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
     to_upgrade = set(pkgs_to_upgrade)
     while True:
         # find smallest set
-        smallest_partition = to_upgrade
+        smallest_partition = set()
         for pkgname in to_upgrade:
             if SIGNAL_STOP_REQUEST:
                 logging.warning("SIGNAL received, stopping")
@@ -467,12 +470,18 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
                 logging.debug("found leaf package %s" % pkg.name)
                 smallest_partition = changes
                 break
-            if len(changes) < len(smallest_partition):
+            if (len(changes) < len(smallest_partition) or
+                    len(smallest_partition) == 0):
                 logging.debug("found partition of size %s (%s)" % (
                     len(changes), changes))
                 smallest_partition = changes
             cache.clear()
 
+        if len(smallest_partition) == 0:
+            logging.info(
+                _("Some packages can not be upgraded: %s") % (to_upgrade))
+            break
+
         # write progress log information
         if len(pkgs_to_upgrade) > 0:
             percent = ((len(pkgs_to_upgrade) - len(to_upgrade)) /
@@ -486,6 +495,14 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, blacklist,
         logging.debug("applying set %s" % smallest_partition)
         rewind_cache(cache, [cache[name] for name in smallest_partition])
 
+        # one last check, just to be sure
+        if not check_changes_for_sanity(cache, allowed_origins, blacklist,
+                                        whitelist):
+            logging.error("Failed sanity check while attempting to commit: "
+                          "this should never happen")
+            cache.clear()
+            return False
+
         res, error = cache_commit(cache, logfile_dpkg, verbose, install_log)
         if error:
             if verbose:
@@ -559,8 +576,8 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist):
             # dependencies
             strict_whitelist = apt_pkg.config.find_b(
                 "Unattended-Upgrade::Package-Whitelist-Strict", False)
-            if (strict_whitelist and not
-                    is_pkgname_in_whitelist(pkg.name, whitelist)):
+            if (strict_whitelist and
+                    not is_pkgname_in_whitelist(pkg.name, whitelist)):
 
                 logging.debug("pkg '%s' package is not whitelisted" %
                               pkg.name)
@@ -749,7 +766,7 @@ def setup_apt_listchanges(conf="/etc/apt/listchanges.conf"):
         os.environ["APT_LISTCHANGES_FRONTEND"] = "none"
 
 
-def _send_mail_using_mailx(to_address, subject, body):
+def _send_mail_using_mailx(from_address, to_address, subject, body):
     # ensure that the body is a byte stream and that we do not
     # break on encoding errors (the default error mode is "strict")
     encoded_body = body.encode(
@@ -758,7 +775,7 @@ def _send_mail_using_mailx(to_address, subject, body):
     # unicode encoding errors (e.g. because the user is running a
     # ascii only system like the buildds)
     mail = subprocess.Popen(
-        [MAIL_BINARY, "-s", subject, to_address],
+        [MAIL_BINARY, "-r", from_address, "-s", subject, to_address],
         stdin=subprocess.PIPE, universal_newlines=False)
     mail.stdin.write(encoded_body)
     mail.stdin.close()
@@ -766,10 +783,11 @@ def _send_mail_using_mailx(to_address, subject, body):
     return ret
 
 
-def _send_mail_using_sendmail(to_address, subject, body):
+def _send_mail_using_sendmail(from_address, to_address, subject, body):
     # format as a proper mail
     msg = Message()
     msg['Subject'] = subject
+    msg['From'] = from_address
     msg['To'] = to_address
     # order is important here, Message() first, then Charset()
     #  then msg.set_charset()
@@ -805,14 +823,16 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, dpkg_log_content):
             "Unattended-Upgrade::MailOnlyOnError", False)):
         return
     # Check if reboot-required flag is present
+    reboot_flag_str = _(
+        "[reboot required]") if os.path.isfile(REBOOT_REQUIRED_FILE) else ""
+    # Check if packages are kept on hold
+    hold_flag_str = _("[package on hold]") if pkgs_kept_back else ""
     logging.debug("Sending mail to '%s'" % to_email)
-    if os.path.isfile(REBOOT_REQUIRED_FILE):
-        subject = _(
-            "[reboot required] unattended-upgrades result for '%s': %s") % (
-            host(), res)
-    else:
-        subject = _("unattended-upgrades result for '%s': '%s'") % (
-            host(), res)
+    subject = _(
+        "{hold_flag}{reboot_flag} unattended-upgrades result for "
+        "'{machine}': {result}").format(
+            hold_flag=hold_flag_str, reboot_flag=reboot_flag_str,
+            machine=host(), result=res).strip()
     body = _("Unattended upgrade returned: %s\n\n") % res
     if os.path.isfile(REBOOT_REQUIRED_FILE):
         body += _(
@@ -835,10 +855,12 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, dpkg_log_content):
     body += _("Unattended-upgrades log:\n")
     body += mem_log.getvalue()
 
+    from_email = apt_pkg.config.find("Unattended-Upgrade::Sender", "root")
+
     if os.path.exists(SENDMAIL_BINARY):
-        ret = _send_mail_using_sendmail(to_email, subject, body)
+        ret = _send_mail_using_sendmail(from_email, to_email, subject, body)
     elif os.path.exists(MAIL_BINARY):
-        ret = _send_mail_using_mailx(to_email, subject, body)
+        ret = _send_mail_using_mailx(from_email, to_email, subject, body)
     else:
         raise AssertionError(
             "This should never be reached, if we are here we either "
@@ -951,6 +973,25 @@ def _setup_logging(options):
     if apt_pkg.config.find("Unattended-Upgrade::Mail", ""):
         mem_log_handler = logging.StreamHandler(mem_log)
         logger.addHandler(mem_log_handler)
+    # Configure syslog if necessary
+    syslogEnable = apt_pkg.config.find_b("Unattended-Upgrade::SyslogEnable",
+                                         False)
+    if syslogEnable:
+        syslogFacility = apt_pkg.config.find(
+            "Unattended-Upgrade::SyslogFacility",
+            "daemon")
+        syslogHandler = logging.handlers.SysLogHandler(
+            address='/dev/log',
+            facility=syslogFacility)
+        syslogHandler.setFormatter(
+            logging.Formatter("unattended-upgrade: %(message)s"))
+        if syslogFacility.lower() in syslogHandler.facility_names.keys():
+            logger.addHandler(syslogHandler)
+            logging.info("Enabled logging to syslog via %s facility "
+                         % syslogFacility)
+        else:
+            logging.warning("Syslog facility %s was not found"
+                            % syslogFacility)
     return mem_log
 
 
@@ -1046,8 +1087,8 @@ def calculate_upgradable_pkgs(cache, options, allowed_origins,
         if options.debug and pkg.is_upgradable:
             logging.debug("Checking: %s (%s)" % (
                 pkg.name, getattr(pkg.candidate, "origins", [])))
-        if (pkg.is_upgradable and not
-                is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
+        if (pkg.is_upgradable and
+                not is_pkgname_in_blacklist(pkg.name, blacklisted_pkgs) and
                 is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs) and
                 is_allowed_origin(pkg.candidate, allowed_origins)):
 
@@ -1066,7 +1107,7 @@ def get_dpkg_log_content(logfile_dpkg, install_start_time):
         logfile_dpkg, install_start_time))
     content = []
     found_start = False
-    with open(logfile_dpkg) as fp:
+    with io.open(logfile_dpkg, encoding='utf-8') as fp:
         # read until we find the last "Log started: "
         for line in fp.readlines():
             # scan for the first entry we need (minimal-step mode
@@ -1086,8 +1127,7 @@ def get_auto_removable(cache):
                 if pkg.is_auto_removable])
 
 
-def do_auto_remove(cache, options, logfile_dpkg, verbose=False):
-    auto_removable = get_auto_removable(cache)
+def do_auto_remove(cache, auto_removable, logfile_dpkg, verbose=False):
     if not auto_removable:
         return True
 
@@ -1095,7 +1135,7 @@ def do_auto_remove(cache, options, logfile_dpkg, verbose=False):
         logging.debug("marking %s for remove" % pkgname)
         cache[pkgname].mark_delete()
     logging.info(_("Packages that are auto removed: '%s'"),
-                 " ".join(auto_removable))
+                 " ".join(sorted(auto_removable)))
 
     # do it
     res, error = cache_commit(cache, logfile_dpkg, verbose)
@@ -1109,6 +1149,41 @@ def do_auto_remove(cache, options, logfile_dpkg, verbose=False):
     return res
 
 
+def clean_downloaded_packages(fetcher):
+    archivedir = os.path.dirname(
+        apt_pkg.config.find_dir("Dir::Cache::archives"))
+    for item in fetcher.items:
+        if os.path.dirname(os.path.abspath(item.destfile)) == archivedir:
+            try:
+                os.unlink(item.destfile)
+            except OSError:
+                pass
+
+
+def is_update_day():
+    # check if patch days are configured
+    patch_days = apt_pkg.config.value_list("Unattended-Upgrade::Update-Days")
+    if not patch_days:
+        return True
+    # validate patch days
+    today = date.today()
+    # abbreviated localized dayname
+    if today.strftime("%a") in patch_days:
+        return True
+    # full localized dayname
+    if today.strftime("%A") in patch_days:
+        return True
+    # by number (Sun: 0, Mon: 1, ...)
+    if today.strftime("%w") in patch_days:
+        return True
+    # today is not a patch day
+    logging.info(
+        "Skipping update check: today is '%s,%s,%s' but patch days are '%s'",
+        today.strftime("%w"), today.strftime("%a"), today.strftime("%A"),
+        patch_days)
+    return False
+
+
 def main(options, rootdir=""):
     # useful for testing
     if rootdir:
@@ -1117,6 +1192,10 @@ def main(options, rootdir=""):
     # setup logging
     mem_log = _setup_logging(options)
 
+    # check if today is a patch day
+    if not is_update_day():
+        return
+
     # format (origin, archive), e.g. ("Ubuntu","dapper-security")
     allowed_origins = get_allowed_origins()
 
@@ -1298,9 +1377,10 @@ def main(options, rootdir=""):
         logging.debug("dpkg is configured not to cause conffile prompts")
 
     # auto-removal
+    previous_autoremovals = get_auto_removable(cache)
     if apt_pkg.config.find_b(
             "Unattended-Upgrade::Remove-Unused-Dependencies", False):
-        pending_autoremovals = get_auto_removable(cache)
+        pending_autoremovals = previous_autoremovals
     else:
         pending_autoremovals = []
 
@@ -1340,7 +1420,7 @@ def main(options, rootdir=""):
     logfile_dpkg = os.path.join(_get_logdir(), 'unattended-upgrades-dpkg.log')
 
     # only perform install step if we actually have packages to install
-    pkg_install_success = None
+    pkg_install_success = True
     shutdown_lock = -1
     if len(pkgs_to_upgrade) > 0:
         # lock for the shutdown check - its fine if the system
@@ -1354,21 +1434,30 @@ def main(options, rootdir=""):
                                          options,
                                          logfile_dpkg)
 
-    # remove unneeded packages if there are any
+    # now check if any auto-removing needs to be done
+    cache = UnattendedUpgradesCache(
+        rootdir=rootdir, allowed_origins=allowed_origins)
+    if cache._depcache.broken_count > 0:
+        print(_("Cache has broken packages, exiting"))
+        logging.error(_("Cache has broken packages, exiting"))
+        sys.exit(1)
+
+    # the user wants *all* auto-removals to be removed
     if apt_pkg.config.find_b(
             "Unattended-Upgrade::Remove-Unused-Dependencies", False):
-        cache = UnattendedUpgradesCache(rootdir=rootdir,
-                                        allowed_origins=allowed_origins)
-        if cache._depcache.broken_count > 0:
-            print(_("Cache has broken packages, exiting"))
-            logging.error(_("Cache has broken packages, exiting"))
-            sys.exit(1)
-
-        remove_success = do_auto_remove(
-            cache, options, logfile_dpkg,
+        auto_removals = get_auto_removable(cache)
+        pkg_install_success &= do_auto_remove(
+            cache, auto_removals, logfile_dpkg,
+            options.verbose or options.debug)
+    # the user wants *only new* auto-removals to be removed
+    elif apt_pkg.config.find_b(
+            "Unattended-Upgrade::Remove-New-Unused-Dependencies", True):
+        # calculate the new auto-removals
+        new_pending_autoremovals = get_auto_removable(cache)
+        auto_removals = new_pending_autoremovals - previous_autoremovals
+        pkg_install_success &= do_auto_remove(
+            cache, auto_removals, logfile_dpkg,
             options.verbose or options.debug)
-        if pkg_install_success is not False:
-            pkg_install_success = remove_success
 
     logging.debug("InstCount=%i DelCount=%i BrokenCount=%i"
                   % (cache._depcache.inst_count,
@@ -1381,6 +1470,13 @@ def main(options, rootdir=""):
         send_summary_mail(
             pkgs, pkg_install_success, pkgs_kept_back, mem_log, log_content)
 
+    # clean after success install (if needed)
+    keep_key = "Unattended-Upgrade::Keep-Debs-After-Install"
+    if (not apt_pkg.config.find_b(keep_key, False) and
+            not options.dry_run and
+            pkg_install_success):
+        clean_downloaded_packages(fetcher)
+
     # FIXME: DRY violation, write_stamp_file() is used above as well
     # write timestamp file
     write_stamp_file()
